import os
import tempfile

from infra_buddy.aws import s3 as s3util
from infra_buddy.aws.cloudformation import CloudFormationBuddy
from infra_buddy.aws.s3 import S3Buddy
from infra_buddy.deploy.deploy import Deploy
from infra_buddy.utility import print_utility


class S3Deploy(Deploy):
    def __init__(self, artifact_id, location, ctx):
        super(S3Deploy, self).__init__(ctx)
        self.location = location
        self.artifact_id = artifact_id
        self.cloud_formation_buddy = CloudFormationBuddy(self.deploy_ctx)

    def _internal_deploy(self, dry_run):
        mkdtemp = tempfile.mkdtemp()
        if not self.artifact_id.endswith(".zip" ):
            self.artifact_id = "{}.zip".format(self.artifact_id)
        artifact_download = "s3://{location}/{artifact_id}".format(location=self.location,artifact_id=self.artifact_id)
        destination_bucket = self.cloud_formation_buddy.get_export_value(param="WWW-Files")
        s3util.download_zip_from_s3_url(artifact_download,destination=mkdtemp)

        to_upload = self.get_filepaths(mkdtemp)
        if dry_run:
            print_utility.banner_warn("Dry Run: Uploading files to - {}".format(destination_bucket),
                                      str(to_upload))
        else:
            split = destination_bucket.split("/")
            if len(split)>1:
                path = "/".join(split[1:])
            else:
                path = ''
            s3 = S3Buddy(self.deploy_ctx, path, split[0])
            print_utility.progress("S3 Deploy: Uploading files to - {}".format(destination_bucket))
            for s3_key, path in to_upload.iteritems():
                print_utility.info("{} - {}".format(destination_bucket, s3_key))
                s3.upload(key_name=s3_key, file=path)

    def get_filepaths(self, local_directory):
        rel_paths = {}
        for root, dirs, files in os.walk(local_directory):
            for filename in files:
                # construct the full local path
                local_path = os.path.join(root, filename)
                # construct the full Dropbox path
                relative_path = os.path.relpath(local_path, local_directory)
                # s3_path = os.path.join(destination, relative_path)
                rel_paths[relative_path] = local_path
        return rel_paths

    def __str__(self):
       return "{} - {}:{}".format(self.__class__.__name__,self.location,self.artifact_id)





import json
import os

import datetime
from jsonschema import validate

from infra_buddy.deploy.cloudformation_deploy import CloudFormationDeploy
from infra_buddy.template.template_manager import TemplateManager
from infra_buddy.utility import print_utility

_SERVICE_DEFINITION_FILE = "service.json"

_MODIFICATIONS = 'service-modifications'

_DEPLOYMENT_PARAMETERS = 'deployment-parameters'
_CI_DEPLOYMENT_PARAMETERS = 'ci-deployment-parameters'
_PROD_DEPLOYMENT_PARAMETERS = 'prod-deployment-parameters'
_DEV_DEPLOYMENT_PARAMETERS = 'dev-deployment-parameters'
_UNIT_TEST_DEPLOYMENT_PARAMETERS = 'unit-test-deployment-parameters'
_DOCKER_REGISTRY = 'registry-url'
_SERVICE_TYPE = 'service-type'
_ROLE = 'role'
_APPLICATION = 'application'


class ServiceDefinition(object):
    schema = {
        "type": "object",
        "properties": {
            _APPLICATION: {"type": "string"},
            _ROLE: {"type": "string"},
            _SERVICE_TYPE: {"type": "string"},
            _DOCKER_REGISTRY: {"type": ["string", "null"], 'maxLength': 1000},
            _DEPLOYMENT_PARAMETERS: {
                "type": "object",
                "properties": {
                },
                "additionalProperties": True
            },
            _CI_DEPLOYMENT_PARAMETERS: {
                "type": "object",
                "properties": {
                },
                "additionalProperties": True
            },
            _PROD_DEPLOYMENT_PARAMETERS: {
                "type": "object",
                "properties": {
                },
                "additionalProperties": True
            },
            _DEV_DEPLOYMENT_PARAMETERS: {
                "type": "object",
                "properties": {
                },
                "additionalProperties": True
            }
        },
        "required": [
            _APPLICATION,
            _ROLE,
            _DEPLOYMENT_PARAMETERS,
            _SERVICE_TYPE
        ]
    }

    def __init__(self, artifact_directory, environment):
        super(ServiceDefinition, self).__init__()
        self.artifact_directory = artifact_directory
        service_definition_path = os.path.join(artifact_directory, _SERVICE_DEFINITION_FILE)
        if not os.path.exists(service_definition_path):
            err_msg = "Service definition ({}) does not exist in artifact directory - {}".format(
                _SERVICE_DEFINITION_FILE,
                artifact_directory)
            print_utility.error(err_msg)
            raise Exception(err_msg)
        with open(service_definition_path, 'r') as fp:
            service_definition = json.load(fp)
            validate(service_definition, self.schema)
            self.application = service_definition[_APPLICATION]
            self.role = service_definition[_ROLE]
            self.service_type = service_definition[_SERVICE_TYPE]
            self.docker_registry = service_definition.get(_DOCKER_REGISTRY,"")
            if _DEPLOYMENT_PARAMETERS in service_definition:
                self.deployment_parameters = service_definition[_DEPLOYMENT_PARAMETERS]
            env_deployment_parameters = '{environment}-deployment-parameters'.format(environment=environment)
            if env_deployment_parameters in service_definition:
                print_utility.info("Updating deployment params with environment"
                                   " specific settings - {}".format(env_deployment_parameters))
                self.deployment_parameters.update(service_definition[env_deployment_parameters])
            self.service_modifications = service_definition.get(_MODIFICATIONS, [])

    def generate_execution_plan(self, template_manager, deploy_ctx):
        # type: (TemplateManager) -> list(Deploy)
        ret = []
        ret.append(CloudFormationDeploy(stack_name=deploy_ctx.stack_name,
                                        template=template_manager.get_known_service(self.service_type),
                                        deploy_ctx=deploy_ctx))
        resource_deploy = template_manager.get_resource_service(self.artifact_directory)
        if resource_deploy:
            ret.append(CloudFormationDeploy(stack_name=deploy_ctx.resource_stack_name,
                                            template=resource_deploy,
                                            deploy_ctx=deploy_ctx))
        else:
            print_utility.info("Addition resource template not located (aws-resources.template).")
        for mod in self.service_modifications:
            template = template_manager.get_known_service_modification(self.service_type,mod)
            ret.append(CloudFormationDeploy(stack_name=deploy_ctx.generate_modification_stack_name(mod),
                                            template=template,
                                            deploy_ctx=deploy_ctx))
        return ret

    @classmethod
    def save_to_file(cls, application,role, deploy_params, service_type,known_service_modifications,destination):
        # type: (str, str, dict,str) -> str
        if destination:
            service_file_path = os.path.join(destination,_SERVICE_DEFINITION_FILE)
            readme_file_path = os.path.join(destination,"README.md")
        else:
            service_file_path = _SERVICE_DEFINITION_FILE
            readme_file_path = "README.md"
        service_definition_object = {
            _APPLICATION: application,
            _ROLE: role,
            _SERVICE_TYPE: service_type,
            _DEPLOYMENT_PARAMETERS: cls._get_params_without_default_values(deploy_params),
            _MODIFICATIONS: []}
        with open(service_file_path, 'w') as def_file:
            json.dump(service_definition_object,def_file)
        with open(readme_file_path,'w') as read:
            read.write("# Service Type: {}\n".format(service_type))
            read.write("Generated on {}\n\n".format(datetime.datetime.now()))
            read.write("Service template may have been modified, please verify usage with:\n")
            read.write(" ```bash\n")
            read.write("infra-buddy validate-template --service-type {}\n".format(service_type))
            read.write("```\n")
            read.write("## Known Service Modifications\n")
            read.write("Define these service modifications in the service.json stanza '{}'\n".format(_MODIFICATIONS))
            read.write(" ```javascript\n")
            read.write("\"{}\":[ \"modification-1\"]'\n".format(_MODIFICATIONS))
            read.write("```\n\n")
            read.write("| Service Modification |\n")
            read.write("| --- |\n")
            for key,val in known_service_modifications.iteritems():
                read.write("| {} |\n".format(key))
            read.write("\n")
            read.write("## Deploy parameters\n\n")
            read.write("Define these parameters in the service.json stanza '{}'\n\n".format(_DEPLOYMENT_PARAMETERS))
            read.write("For environment specific values use the corresponding stanza in the form '<environment>-{}'\n".format(_DEPLOYMENT_PARAMETERS))
            read.write("this is supported for the known environments: 'dev', 'ci', and 'prod'.\n\n")
            if len(deploy_params) > 0:
                read.write("| Parameter | Description | Default Value |\n")
                read.write("| --- | --- | --- |\n")
                for key_, definition in deploy_params.iteritems():
                    if 'default_type' in definition and definition["default_type"] == "property":
                        parameter = definition['key']
                        description = definition.get("description","<None>")
                        default_val = definition.get("default","<None>")
                        read.write("| {} | {} | {} |\n".format(parameter,description,default_val))
        return service_file_path

    @classmethod
    def _get_params_without_default_values(cls, deploy_params):
        ret = {}
        for key_, definition in deploy_params.iteritems():
               if 'default_type' in definition and definition["default_type"] == "property":
                   if 'default_value' not in definition:
                       definition_key_ = definition['key']
                       ret[definition_key_] = os.environ.get(definition_key_,"")
        return ret





import mimetypes
import os
import tempfile
import urlparse
from zipfile import ZipFile

import boto3
import botocore
from boto3.s3.transfer import S3Transfer

from infra_buddy.utility import print_utility


class S3Buddy(object):
    def __init__(self, deploy_ctx, root_path, bucket_name):
        super(S3Buddy, self).__init__()
        self.deploy_ctx = deploy_ctx
        self.s3 = boto3.resource('s3', region_name=self.deploy_ctx.region)
        self.bucket = self.s3.Bucket(bucket_name)
        try:
            self.bucket.create(CreateBucketConfiguration={'LocationConstraint': self.deploy_ctx.region})
        except (botocore.exceptions.ValidationError, botocore.exceptions.ClientError) as err:
            if 'BucketAlreadyOwnedByYou' not in str(err):
                print_utility.info("Error during bucket create - {}".format(str(err)))
        self.bucket_name = bucket_name
        self.deploy_ctx = deploy_ctx
        self.s3 = boto3.resource('s3', region_name=self.deploy_ctx.region)
        self.key_root_path = root_path
        self.url_base = "https://s3-{region}.amazonaws.com/{bucket_name}".format(region=self.deploy_ctx.region,
                                                                                 bucket_name=self.bucket_name)

    def upload(self, file, key_name=None):
        key_name = self._get_upload_bucket_key_name(file, key_name)
        args = {"Key":key_name, "Body":open(file, 'rb')}
        content_type = self._guess_content_type(file)
        if content_type:
            args['ContentType'] = content_type
        self.bucket.put_object(**args)
        print_utility.info("Uploaded file to S3 - Bucket: {} Key: {} Content-Type: {}".format(self.bucket_name,
                                                                                              key_name,
                                                                                              content_type))
        return "{}/{}".format(self.url_base, key_name)

    def _get_upload_bucket_key_name(self, file, key_name=None):
        key_name = (key_name if key_name else os.path.basename(file))
        if self.key_root_path and self.key_root_path is not '':
            return "{path}/{key_name}".format(path=self.key_root_path,
                                              key_name=key_name)
        return key_name

    def get_file_as_string(self, filename):
        obj = self._get_s3_object(filename)
        return obj['Body'].read()

    def _get_s3_object(self, filename):
        key_name = self._get_upload_bucket_key_name(file=None, key_name=filename)
        obj = self.s3.meta.client.get_object(Bucket=self.bucket_name, Key=key_name)
        return obj

    def _guess_content_type(self,filename):
        """Given a filename, guess it's content type.
        If the type cannot be guessed, a value of None is returned.
        """
        try:
            return mimetypes.guess_type(filename)[0]
        # This catches a bug in the mimetype libary where some MIME types
        # specifically on windows machines cause a UnicodeDecodeError
        # because the MIME type in the Windows registery has an encoding
        # that cannot be properly encoded using the default system encoding.
        # https://bugs.python.org/issue9291
        #
        # So instead of hard failing, just log the issue and fall back to the
        # default guessed content type of None.
        except UnicodeDecodeError:
            LOGGER.debug(
                'Unable to guess content type for %s due to '
                'UnicodeDecodeError: ', filename, exc_info=True
            )



class CloudFormationDeployS3Buddy(S3Buddy):
    def __init__(self, deploy_ctx, ):
        super(CloudFormationDeployS3Buddy, self).__init__(deploy_ctx=deploy_ctx,
                                                          root_path=deploy_ctx.cf_deploy_resource_path,
                                                          bucket_name=deploy_ctx.cf_bucket_name)


def download_zip_from_s3_url(s3_url, destination):
    # type: (str, dest_directory) -> None
    parsed = urlparse.urlparse(s3_url)
    bucket = parsed.hostname
    key = parsed.path[1:]  # strip leading /
    s3 = boto3.resource('s3')
    with tempfile.NamedTemporaryFile() as temporary_file:
        temp_file_path = temporary_file.name
    print_utility.info("Downloading zip from s3: {} - {}:{}".format(s3_url, key, temp_file_path))
    s3.Bucket(bucket).download_file(key, temp_file_path)
    with ZipFile(temp_file_path) as zf:
        zf.extractall(destination)



#!/usr/bin/env python

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'infra-buddy',
        version = '0.1.92',
        description = 'CLI for deploying micro-services',
        long_description = 'CLI for deploying micro-services',
        author = '',
        author_email = '',
        license = 'Apache 2.0',
        url = 'https://github.com/AlienVault-Engineering/infra-buddy',
        scripts = ['scripts/infra-buddy'],
        packages = [
            'infra_buddy',
            'infra_buddy.aws',
            'infra_buddy.commands',
            'infra_buddy.context',
            'infra_buddy.deploy',
            'infra_buddy.notifier',
            'infra_buddy.template',
            'infra_buddy.utility',
            'infra_buddy.commands.bootstrap',
            'infra_buddy.commands.deploy_cloudformation',
            'infra_buddy.commands.deploy_service',
            'infra_buddy.commands.generate_artifact_manifest',
            'infra_buddy.commands.generate_service_definition',
            'infra_buddy.commands.introspect',
            'infra_buddy.commands.validate_template'
        ],
        namespace_packages = [],
        py_modules = [],
        classifiers = [
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python'
        ],
        entry_points = {},
        data_files = [],
        package_data = {
            'infra_buddy': ['template/builtin-templates.json']
        },
        install_requires = [
            'click',
            'boto3',
            'pydash',
            'jsonschema',
            'requests',
            'datadog'
        ],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        keywords = '',
        python_requires = '',
        obsoletes = [],
    )





class Deploy(object):


    
    def __init__(self,deploy_ctx):
        super(Deploy, self).__init__()
        self.deploy_ctx = deploy_ctx
        self.stack_name = None

    def do_deploy(self,dry_run=False):
        self.deploy_ctx.push_deploy_ctx(self)
        try:
            return self._internal_deploy(dry_run)
        finally:
            self.deploy_ctx.pop_deploy_ctx()

    def _internal_deploy(self,dry_run):
        pass

    def __str__(self):
        return "{} - {}".format(self.__class__.__name__,self.stack_name)

